/*  blddecli.c - implementation of DDE CLIent classes.
 *  
 *  TDde - is a local class that maintains the DDEML instance variable.
 *  TDdeConversation
 *      - is the concrete version of the abstract ADdeConversation
 *        exposed in ddecli.h.
 */

#include <assert.h>
#include <limits.h>
#include <string.h>
#include <windows.h>
#include <ddeml.h>

#include "blddecli.h"

// Declare DDE callback function
HDDEDATA    EXPENTRY    DdeCallback(WORD TransactionType,
    WORD ClipboardFormat, HCONV ConversationHandle,
    HSZ String1, HSZ String2, HDDEDATA MemoryHandle,
    DWORD Data1, DWORD Data2);

//  TDde - local class to maintain global DDE data.
////////////////////////////////////////////////////////////
class   TDde
    {
public:
    TDde() {};
   ~TDde();
    int             Initialize();
    int             Uninitialize();
    int             LastError();
    static DWORD    Instance;
    static int      ErrorCode;
    };
DWORD   TDde::Instance  = 0;
int     TDde::ErrorCode = 0;

TDde::~TDde()
    {
    if(TDde::Instance)  // if DDE instance is initialized
        {
        DdeUninitialize(TDde::Instance);
        TDde::Instance  = 0;
        }
    }

int     TDde::Initialize()
    {
    if(TDde::Instance != NULL)
        return TRUE;
    else if(DdeInitialize(&TDde::Instance,
            (PFNCALLBACK) DdeCallback,
            APPCMD_CLIENTONLY, 0L) == DMLERR_NO_ERROR)
        return TRUE;
    else
        return FALSE;
    }

int     TDde::Uninitialize()
    {
    int     Status = TRUE;

    if(TDde::Instance != NULL)
        {
        Status  = DdeUninitialize(TDde::Instance);
        TDde::Instance  = 0;
        }
    return Status;
    }

int     TDde::LastError()
    {
    int Result = (int)DdeGetLastError(TDde::Instance);
    return Result;
    }

static TDde    DDE;    // static destructor handles de-init


// TDdeString - Automate HSZ-string conversion
////////////////////////////////////////////////////////////
class   TDdeString
    {
public  :
    TDdeString(const char *String);
   ~TDdeString();
    operator HSZ() { return StringHandle; }
private :
    HSZ     StringHandle;
    };

TDdeString::TDdeString(const char *String)
    {
    StringHandle    = DdeCreateStringHandle(
        TDde::Instance, String, CP_WINANSI);
    assert(StringHandle != NULL);
    }

TDdeString::~TDdeString()
    {
    DdeFreeStringHandle(TDde::Instance, StringHandle);
    }

// don't know why Borland requires this
ADdeConversation::~ADdeConversation()
    {
    }


// TDdeConversation - implementation for abstract DDE conv.
////////////////////////////////////////////////////////////
class   TDdeConversation : public ADdeConversation
    {
public:
    TDdeConversation();
    virtual ~TDdeConversation();
    int     Initialize(const char *Server, const char *Topic);
    virtual int Request(const char *Item, char *Buffer,
                                               size_t Size);
    virtual int Execute(const char *Command);
    virtual int Poke(const char *Item, const char *Command);
private:
    HCONV       ConversationHandle_;
    };

TDdeConversation::TDdeConversation()
    {
    ConversationHandle_ = 0;
    }

TDdeConversation::~TDdeConversation()
    {
    if(ConversationHandle_)
        {
        DdeDisconnect(ConversationHandle_);
        ConversationHandle_ = 0;
        }
    }

int     TDdeConversation::Initialize(const char *Server,
                                          const char *Topic)
    {
    TDdeString  DdeServer(Server);
    TDdeString  DdeTopic(Topic);
    ConversationHandle_   = DdeConnect(TDde::Instance,
                    DdeServer, DdeTopic, (PCONVCONTEXT)NULL);
    return ConversationHandle_ != 0;
    }

int    TDdeConversation::Execute(const char *Command)
    {
    DWORD       Status;
    HDDEDATA    Result;

    Result  = DdeClientTransaction((LPBYTE)Command, strlen(Command),
                ConversationHandle_, NULL, CF_TEXT,
                XTYP_EXECUTE, 5000, &Status);
    if(int(Result) == TRUE)
        return 0;
    else if(Result == FALSE)
        {
        return int(Status & 0x00FF);
        }
    else
        return -1;
    }

int    TDdeConversation::Poke(const char *Item, const char *Command)
    {
    DWORD       Status;
    TDdeString  DdeItem(Item);
    HDDEDATA    Result;

    Result  = DdeClientTransaction((LPBYTE)Command, strlen(Command),
                ConversationHandle_, DdeItem, CF_TEXT,
                XTYP_POKE, 5000, &Status);
    if(int(Result) == TRUE)
        return 0;
    else if(Result == FALSE)
        {
        return int(Status & 0x00FF);
        }
    else
        return -1;
    }

int TDdeConversation::Request(const char *Item,
                                  char *Buffer, size_t Size)
    {
    DWORD       Status;
    TDdeString  DdeItem(Item);
    HDDEDATA    Result;

    if(Size == 0)   // if caller places no size restriction
        Size    = UINT_MAX;
    --Size;     // because we will NULL-terminate
    Result  = DdeClientTransaction((LPBYTE)NULL, 0,
                ConversationHandle_, DdeItem, CF_TEXT,
                XTYP_REQUEST, 5000, &Status);
    if(Result)
        {
        DWORD   DataLength;
        BYTE FAR *Data = DdeAccessData(Result, &DataLength);
        if(Data)
            {
            assert((DataLength&0xFFFF0000L) == 0);
            if((size_t)DataLength < Size)
                Size    = (size_t)DataLength;
            memcpy(Buffer, Data, Size);
            Buffer[Size]    = '\0';
            DdeUnaccessData(Result);
            }
        DdeFreeDataHandle(Result);
        return 0;
        }
    else
        return int(Status & 0x00FF);
    }


#ifdef __BORLANDC__
    #pragma argsused
#endif
HDDEDATA    EXPENTRY    DdeCallback(WORD TransactionType,
    WORD ClipboardFormat, HCONV ConversationHandle,
    HSZ String1, HSZ String2, HDDEDATA MemoryHandle,
    DWORD Data1, DWORD Data2)
    {
    switch(TransactionType)
        {
        case    XTYP_REGISTER   :
        case    XTYP_UNREGISTER :
            return (HDDEDATA)NULL;
        case    XTYP_ADVDATA    :
            return (HDDEDATA) DDE_FACK;
        case    XTYP_XACT_COMPLETE  :
            return (HDDEDATA)NULL;
        case    XTYP_DISCONNECT :
            return (HDDEDATA)NULL;
        }
    return (HDDEDATA)NULL;
    }


//  TDdeClient member functions
////////////////////////////////////////////////////////////

ADdeConversation *TDdeClient::NewConversation(
                      const char *Server, const char *Topic,
                                const char *CommandLine)
    {
    DDE.Initialize();

    TDdeConversation *Conversation = new TDdeConversation();

    if(Conversation->Initialize(Server, Topic))
        return Conversation;
    else
        {
        delete Conversation;
        if(  CommandLine != NULL
          && LastError() == 0
          )
            {
            WinExec(CommandLine, SW_HIDE);
            return NewConversation(Server, Topic);
            }
        return 0;
        }
    }

int TDdeClient::Initialize()
    {
    return DDE.Initialize();
    }

int TDdeClient::LastError()
    {
    return DDE.LastError();
    }


